;; Generated package description from el-job.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "el-job" "2.5.2.0.20251006.162915" "Contrived way to call a function using all CPU cores" '((emacs "29.1")) :commit "496063123f332b39b2ef6613dcb62809da89cc1b" :authors '(("Martin Edström" . "meedstrom@runbox.eu")) :maintainer '("Martin Edström" . "meedstrom@runbox.eu") :keywords '("processes") :url "https://github.com/meedstrom/el-job")
